/*
 * [Pluck.java]
 *
 * Summary: extracts data matching a Regex pattern from a collection of files.
 *
 * Copyright: (c) 2008-2017 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.8+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2008-06-04 initial release
 */
package com.mindprod.pluck;

import com.mindprod.commandline.CommandLine;
import com.mindprod.common18.EIO;
import com.mindprod.csv.CSVWriter;
import com.mindprod.filter.AllButSVNDirectoriesFilter;
import com.mindprod.filter.ExtensionListFilter;
import com.mindprod.hunkio.HunkIO;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static java.lang.System.*;

/**
 * extracts data matching a Regex pattern from a collection of files.
 * <p/>
 * Replaced by extract, extracti, without, withouti.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.0 2008-06-03 initial version
 * @since 2008-06-04
 */
public class Pluck
    {
    private static final int FIRST_COPYRIGHT_YEAR = 2008;

    /**
     * undisplayed copyright notice
     */
    @SuppressWarnings( { "UnusedDeclaration" } )
    private static final String EMBEDDED_COPYRIGHT =
            "Copyright: (c) 2008-2017 Roedy Green, Canadian Mind Products, http://mindprod.com";

    /**
     * date this version was released.
     */
    @SuppressWarnings( { "UnusedDeclaration" } )
    private static final String RELEASE_DATE = "2008-06-03";

    /**
     * how to use the command line
     */
    private static final String USAGE = "\nPluck requires (regex) (outputfile) (input dirs and files)";

    /**
     * embedded version string.
     */
    @SuppressWarnings( { "UnusedDeclaration" } )
    private static final String VERSION_STRING = "1.0";

    /**
     * the regex pattern we scan for
     */
    private static Pattern pattern;

    /**
     * where we write our findings from the regex scans.
     */
    @SuppressWarnings( { "FieldCanBeLocal" } )
    private static CSVWriter w;

    /**
     * true if want progress messages on the console
     */
    private static boolean wantProgressMessages;

    /**
     * compacts HTML files.
     *
     * @param args regex, output file, then names of files to process, dirs, files, -s, *.*, no wildcards.
     */
    public static void main( String[] args )
        {
        if ( args.length < 3 )
            {
            throw new IllegalArgumentException( USAGE );
            }
        // will throw an detailed exception if this is not a valid string.
        pattern = Pattern.compile( args[ 0 ] );
        out.println( "regex pattern: {" + args[ 0 ] + "}" );
        final String output = args[ 1 ];
        out.println( "output file: " + output );
        try
            {
            if ( output.equalsIgnoreCase( "console" ) )
                {
                w = new CSVWriter( new PrintWriter( new BufferedWriter( new OutputStreamWriter( out ), 4 * 1024 ), true ) );
                wantProgressMessages = false;
                }
            else
                {
                w = new CSVWriter( EIO.getPrintWriter( new File( output ), 64 * 1024, EIO.UTF8 ) );
                wantProgressMessages = true;
                }
            }
        catch ( IOException e )
            {
            throw new IllegalArgumentException( "Pluck cannot open " + output + " for output." );
            }
        // trim off regex and output file, leaving just list of dirs, files and flags.
        args[ 0 ] = null;
        args[ 1 ] = null;
        out.println( "Gathering files to scan for regex matches..." );
        CommandLine commandLine = new CommandLine( args,
                new AllButSVNDirectoriesFilter(),
                new ExtensionListFilter( ExtensionListFilter.COMMON_TEXT_EXTENSIONS ) );
        if ( commandLine.size() == 0 )
            {
            throw new IllegalArgumentException( "No files found to process\n" + USAGE );
            }
        for ( File file : commandLine )
            {
            try
                {
                scanFile( file );
                }
            catch ( FileNotFoundException e )
                {
                err.println( "Error: "
                             + EIO.getCanOrAbsPath( file )
                             + " not found." );
                // keep going
                }
            catch ( Exception e )
                {
                err.println();
                e.printStackTrace( err );
                err.println( " in file " + EIO.getCanOrAbsPath( file ) );
                err.println();
                // keep going
                }
            } // end for
        w.close();
        } // end main

    /**
     * Scan file for all occurences of the regex pattern.  Output a CSV line for each hit, giving
     * the parsed group fields, including group(0) for the whole pattern.
     *
     * @param fileBeingProcessed the file currently being processed.
     *
     * @throws java.io.IOException Suppress IntelliJ Code Analyse that wants to make this private.
     * @noinspection WeakerAccess, SameParameterValue, StringEquality
     */
    public static void scanFile( File fileBeingProcessed ) throws IOException
        {
        if ( wantProgressMessages )
            {
            out.print( "Scanning " + fileBeingProcessed + " " );
            }
        String big = HunkIO.readEntireFile( fileBeingProcessed );
        final Matcher m = pattern.matcher( big );
        while ( m.find() )
            {
            if ( wantProgressMessages )
                {
                out.print( "." );
                }
            w.put( fileBeingProcessed.toString() );
            final int gc = m.groupCount();
            if ( gc == 0 )
                {
                // no subgroups, emit only the whole thing
                w.put( m.group( 0 ) );
                }
            else
                {
                // We have subgroups, emit each one. We don't want the whole pattern.
                // loops runs from from 1 to gc, not 0 to gc-1 as is traditional.
                for ( int i = 1; i <= gc; i++ )
                    {
                    w.put( m.group( i ) );
                    }
                }
            w.nl();
            }
        if ( wantProgressMessages )
            {
            out.println();
            }
        } // end scanFile
    }
